from abc import ABC, abstractmethod
from typing import Any, Dict

class CloneableFile(ABC):
    """
    (en) An abstract class that has deep copy and serialization capabilities
    and is managed by the FileStateManager.

    (ja) ディープコピーとシリアライズの機能を持ち、FileStateManagerで管理される抽象クラス。
    """

    def __init__(self):
        """Normal constructor."""
        pass

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> 'CloneableFile':
        """
        (en) Restore this object from the dictionary.
        (ja) このオブジェクトを辞書から復元します。

        Args:
            src (Dict[str, Any]): A dictionary made with to_dict of this class.

        Returns:
            CloneableFile: 復元されたオブジェクト
        """
        raise NotImplementedError("from_dict must be implemented in subclasses")

    @abstractmethod
    def clone(self) -> 'CloneableFile':
        """
        (en) Returns a deep copy of this object.
        (ja) このオブジェクトのディープコピーを返します。
        """
        pass

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """
        (en) Convert the object to a dictionary.
        The returned dictionary can only contain primitive types, null, lists
        or maps with only primitive elements.
        If you want to include other classes,
        the target class should inherit from this class and chain calls to_dict.

        (ja) このオブジェクトを辞書に変換します。
        戻り値の辞書にはプリミティブ型かプリミティブ型要素のみのリスト
        またはマップ等、そしてNoneのみを含められます。
        それ以外のクラスを含めたい場合、対象のクラスもこのクラスを継承し、
        to_dictを連鎖的に呼び出すようにしてください。
        """
        pass



import unittest
from file_state_manager.file_state_manager_base import FileStateManager
from file_state_manager.cloneable_file import CloneableFile
from file_state_manager.util_object_hash import UtilObjectHash


# -----------------------------
# ExampleClass 定義
# -----------------------------
class ExampleClassChild(CloneableFile):
    def __init__(self, message: str):
        super().__init__()
        self.message = message

    @classmethod
    def from_dict(cls, src: dict):
        return cls(src["message"])

    def clone(self):
        return ExampleClassChild.from_dict(self.to_dict())

    def to_dict(self):
        return {"message": self.message}

    def __eq__(self, other):
        return isinstance(other, ExampleClassChild) and self.message == other.message

    def __hash__(self):
        return hash(self.message)


class ExampleClass(CloneableFile):
    def __init__(self, count: int, child: ExampleClassChild):
        super().__init__()
        self.count = count
        self.child = child

    @classmethod
    def from_dict(cls, src: dict):
        return cls(src["count"], ExampleClassChild.from_dict(src["child"]))

    def clone(self):
        return ExampleClass.from_dict(self.to_dict())

    def to_dict(self):
        return {"count": self.count, "child": self.child.to_dict()}

    def __eq__(self, other):
        return isinstance(other, ExampleClass) and self.count == other.count and self.child == other.child

    def __hash__(self):
        return hash((self.count, self.child))


# -----------------------------
# unittest
# -----------------------------
class TestFileStateManager(unittest.TestCase):
    def test_undo_redo(self):
        save_file = ExampleClass(0, ExampleClassChild("First State"))
        fsm = FileStateManager(save_file, stack_size=None)
        self.assertFalse(fsm.can_undo())

        save_file.child.message = "Second State"
        fsm.push(save_file)
        self.assertTrue(fsm.can_undo())
        self.assertFalse(fsm.can_redo())
        self.assertIsNone(fsm.redo())
        self.assertEqual(fsm.undo().child.message, "First State")
        self.assertTrue(fsm.can_redo())
        self.assertEqual(fsm.redo().child.message, "Second State")

        fsm.undo()
        save_file.child.message = "Third State"
        fsm.push(save_file)
        self.assertTrue(fsm.can_undo())
        self.assertEqual(fsm.undo().child.message, "First State")
        self.assertTrue(fsm.can_redo())
        self.assertEqual(fsm.redo().child.message, "Third State")
        fsm.undo()
        self.assertFalse(fsm.can_undo())
        self.assertIsNone(fsm.undo())

    def test_stack_size(self):
        save_file = ExampleClass(0, ExampleClassChild("First State"))
        fsm = FileStateManager(save_file, stack_size=2)
        save_file.child.message = "Second State"
        fsm.push(save_file)
        self.assertTrue(fsm.can_undo())
        self.assertEqual(fsm.undo().child.message, "First State")
        self.assertTrue(fsm.can_redo())
        self.assertEqual(fsm.redo().child.message, "Second State")
        save_file.child.message = "Third State"
        fsm.push(save_file)
        self.assertTrue(fsm.can_undo())
        self.assertEqual(fsm.undo().child.message, "Second State")
        self.assertTrue(fsm.can_redo())
        self.assertEqual(fsm.redo().child.message, "Third State")
        self.assertTrue(fsm.can_undo())
        self.assertEqual(fsm.undo().child.message, "Second State")
        self.assertFalse(fsm.can_undo())

    def test_history_restore(self):
        save_file = ExampleClass(0, ExampleClassChild("First State"))
        fsm = FileStateManager(save_file, stack_size=30)
        save_file.child.message = "Second State"
        fsm.push(save_file)

        # Save and restore history
        history = [i.to_dict() for i in fsm.get_stack()]
        restored_history = [ExampleClass.from_dict(i) for i in history]
        restored_fsm = FileStateManager(restored_history.pop(0), stack_size=30)
        for i in restored_history:
            restored_fsm.push(i)

        restored_now_state = restored_fsm.now()
        self.assertEqual(restored_now_state.child.message, "Second State")
        self.assertTrue(restored_fsm.can_undo())
        if restored_fsm.can_undo():
            self.assertEqual(restored_fsm.undo().child.message, "First State")

    def test_enable_diff_check(self):
        # non-enabled test
        save_file1 = ExampleClass(0, ExampleClassChild("First State"))
        save_file1_clone = save_file1.clone()
        self.assertEqual(hash(save_file1), hash(save_file1_clone))
        self.assertEqual(save_file1, save_file1_clone)

        fsm1 = FileStateManager(save_file1, stack_size=30)
        save_file1.child.message = "Second State"
        self.assertNotEqual(hash(save_file1), hash(save_file1_clone))
        self.assertNotEqual(save_file1, save_file1_clone)
        fsm1.push(save_file1)
        pre_index1 = fsm1.now_index()
        fsm1.push(save_file1)
        self.assertNotEqual(pre_index1, fsm1.now_index())
        save_file1.child.message = "Third State"
        fsm1.push(save_file1)
        self.assertNotEqual(pre_index1, fsm1.now_index())

        # enabled test
        save_file2 = ExampleClass(0, ExampleClassChild("First State"))
        fsm2 = FileStateManager(save_file2, stack_size=30, enable_diff_check=True)
        save_file2.child.message = "Second State"
        fsm2.push(save_file2)
        pre_index2 = fsm2.now_index()
        fsm2.push(save_file2)
        self.assertEqual(pre_index2, fsm2.now_index())
        save_file2.child.message = "Third State"
        fsm2.push(save_file2)
        self.assertNotEqual(pre_index2, fsm2.now_index())

        self.assertEqual(fsm2.undo().child.message, "Second State")
        self.assertEqual(fsm2.redo().child.message, "Third State")

    def test_skip_next_push(self):
        save_file1 = ExampleClass(0, ExampleClassChild("First State"))
        fsm1 = FileStateManager(save_file1, stack_size=30)
        self.assertEqual(fsm1.now_index(), 0)
        save_file1.child.message = "Second State"
        fsm1.push(save_file1)
        self.assertEqual(fsm1.now_index(), 1)
        save_file1.child.message = "Third State"
        fsm1.skip_next_push()
        fsm1.push(save_file1)
        self.assertEqual(fsm1.now_index(), 1)
        fsm1.push(save_file1)
        self.assertEqual(fsm1.now_index(), 2)


class TestUtilObjectHash(unittest.TestCase):
    def test_util_object_hash(self):
        m1 = {"a": "a"}
        m2 = {"a": "a"}
        m3 = {"a": "b"}
        m4 = {"b": "a"}
        m5 = {"a": "a", "b": "b"}

        self.assertEqual(UtilObjectHash.calc_map(m1), UtilObjectHash.calc_map(m2))
        self.assertNotEqual(UtilObjectHash.calc_map(m1), UtilObjectHash.calc_map(m3))
        self.assertNotEqual(UtilObjectHash.calc_map(m1), UtilObjectHash.calc_map(m4))
        self.assertNotEqual(UtilObjectHash.calc_map(m1), UtilObjectHash.calc_map(m5))

        m6 = {"a": {"a": 1}}
        m7 = {"a": {"a": 1}}
        m8 = {"a": {"a": 2}}
        m9 = {"b": {"a": 1}}
        m10 = {"a": {"a": 1}, "b": {"a": 1}}
        self.assertEqual(UtilObjectHash.calc_map(m6), UtilObjectHash.calc_map(m7))
        self.assertNotEqual(UtilObjectHash.calc_map(m6), UtilObjectHash.calc_map(m8))
        self.assertNotEqual(UtilObjectHash.calc_map(m6), UtilObjectHash.calc_map(m9))
        self.assertNotEqual(UtilObjectHash.calc_map(m6), UtilObjectHash.calc_map(m10))

        l1 = ["a"]
        l2 = ["a"]
        l3 = ["b"]
        l4 = ["a", "b"]
        self.assertEqual(UtilObjectHash.calc_list(l1), UtilObjectHash.calc_list(l2))
        self.assertNotEqual(UtilObjectHash.calc_list(l1), UtilObjectHash.calc_list(l3))
        self.assertNotEqual(UtilObjectHash.calc_list(l1), UtilObjectHash.calc_list(l4))

        l6 = [[1]]
        l7 = [[1]]
        l8 = [[2]]
        l9 = [[1], [1]]
        self.assertEqual(UtilObjectHash.calc_list(l6), UtilObjectHash.calc_list(l7))
        self.assertNotEqual(UtilObjectHash.calc_list(l6), UtilObjectHash.calc_list(l8))
        self.assertNotEqual(UtilObjectHash.calc_list(l6), UtilObjectHash.calc_list(l9))

        s1 = {"a"}
        s2 = {"a"}
        s3 = {"b"}
        s4 = {"a", "b"}
        self.assertEqual(UtilObjectHash.calc_set(s1), UtilObjectHash.calc_set(s2))
        self.assertNotEqual(UtilObjectHash.calc_set(s1), UtilObjectHash.calc_set(s3))
        self.assertNotEqual(UtilObjectHash.calc_set(s1), UtilObjectHash.calc_set(s4))

        s6 = {frozenset({1})}
        s7 = {frozenset({1})}
        s8 = {frozenset({2})}
        s9 = {frozenset({1}), frozenset({2})}
        self.assertEqual(UtilObjectHash.calc_set(s6), UtilObjectHash.calc_set(s7))
        self.assertNotEqual(UtilObjectHash.calc_set(s6), UtilObjectHash.calc_set(s8))
        self.assertNotEqual(UtilObjectHash.calc_set(s6), UtilObjectHash.calc_set(s9))


if __name__ == "__main__":
    unittest.main()



from typing import Any, Dict, List, Set


class UtilObjectHash:
    """
    (en) This is a utility for object hash calculations.
    This makes it easy to calculate hashes, for example
    if you want to enable enableDiffCheck flag in FileStateManager.

    (ja) これはオブジェクトハッシュ計算用のユーティリティです。
    利用することで、FileStateManagerのenableDiffCheckフラグを有効化したい場合などに、
    ハッシュ計算を簡単に行えます。
    """

    @staticmethod
    def calc_map(m: Dict[Any, Any]) -> int:
        """
        (en) Calculate hash code for map. Supports nesting of Maps, Lists, and Sets.
        (ja) Mapのハッシュコードを計算します。Map, List, Setのネストに対応しています。
        """
        r = 17
        for key, value in m.items():
            if isinstance(value, dict):
                r = 37 * r + hash(key)
                r = 37 * r + UtilObjectHash.calc_map(value)
            elif isinstance(value, list):
                r = 37 * r + hash(key)
                r = 37 * r + UtilObjectHash.calc_list(value)
            elif isinstance(value, set):
                r = 37 * r + hash(key)
                r = 37 * r + UtilObjectHash.calc_set(value)
            else:
                r = 37 * r + hash(key)
                r = 37 * r + (hash(value) if value is not None else 0)
        return r

    @staticmethod
    def calc_list(lst: List[Any]) -> int:
        """
        (en) Calculate hash code for list. Supports nesting of Maps, Lists, and Sets.
        (ja) Listのハッシュコードを計算します。Map, List, Setのネストに対応しています。
        """
        r = 17
        for i, v in enumerate(lst):
            if isinstance(v, dict):
                r = 37 * r + (UtilObjectHash.calc_map(v) ^ i)
            elif isinstance(v, list):
                r = 37 * r + (UtilObjectHash.calc_list(v) ^ i)
            elif isinstance(v, set):
                r = 37 * r + (UtilObjectHash.calc_set(v) ^ i)
            else:
                r = 37 * r + ((hash(v) if v is not None else 0) ^ i)
        return r

    @staticmethod
    def calc_set(s: Set[Any]) -> int:
        """
        (en) Calculate hash code for set. Supports nesting of Maps, Lists, and Sets.
        (ja) Setのハッシュコードを計算します。Map, List, Setのネストに対応しています。
        """
        r = 17
        for v in s:
            if isinstance(v, dict):
                r = 37 * r + UtilObjectHash.calc_map(v)
            elif isinstance(v, list):
                r = 37 * r + UtilObjectHash.calc_list(v)
            elif isinstance(v, set):
                r = 37 * r + UtilObjectHash.calc_set(v)
            else:
                r = 37 * r + (hash(v) if v is not None else 0)
        return r



from typing import List, Optional
from .cloneable_file import CloneableFile


class FileStateManager:
    """
    (en) This is a manager class that allows you to Undo and Redo
    complex file changes.
    (ja) 複雑なファイルの変更をUndo,Redoできるようにするためのマネージャクラスです。
    """

    def __init__(self, f: CloneableFile, stack_size: Optional[int] = None, enable_diff_check: bool = False):
        """
        Args:
            f (CloneableFile): A managed file. Can be new or loaded.
            stack_size (Optional[int]): Maximum stack size. None for unlimited.
            enable_diff_check (bool): Whether to skip unchanged pushes (requires __eq__ and __hash__ in files)
        """
        self._ur_stack: List[CloneableFile] = []
        self._now_index: int = -1
        self.stack_size: Optional[int] = stack_size
        self.enable_diff_check: bool = enable_diff_check
        self._skip_next: bool = False

        self.push(f)

    def push(self, f: CloneableFile):
        """
        (en) Adds elements that have been changed.
        (ja) 変更が加えられた要素を追加します。
        """
        if self._skip_next:
            self._skip_next = False
            return

        if self.enable_diff_check and self._ur_stack:
            if self._ur_stack[self._now_index] == f:
                return

        cloned_data = f.clone()

        # スタック内に収まるように調整
        if self.stack_size is None:
            self._now_index += 1
        else:
            if self._now_index < self.stack_size - 1:
                self._now_index += 1
            else:
                self._ur_stack.pop(0)

        # 新しいデータを追加し、以降のデータがあれば削除
        if self._now_index < len(self._ur_stack):
            self._ur_stack[self._now_index] = cloned_data
            # 余分なデータを削除
            del self._ur_stack[self._now_index + 1:]
        else:
            self._ur_stack.append(cloned_data)

    def skip_next_push(self):
        """
        (en) Calling this will disable the next push.
        (ja) これを呼び出すと、次回のpushが無効化されます。
        """
        self._skip_next = True

    def can_undo(self) -> bool:
        """
        (en) Returns true only if Undo is possible.
        (ja) Undo可能な場合のみtrueを返します。
        """
        return len(self._ur_stack) >= 2 and self._now_index >= 1

    def can_redo(self) -> bool:
        """
        (en) Returns true only if Redo is possible.
        (ja) Redo可能な場合のみtrueを返します。
        """
        return self._now_index < len(self._ur_stack) - 1

    def undo(self) -> Optional[CloneableFile]:
        """
        (en) Returns the previously pushed data.
        (ja) １つ前にpushされたデータを返します。
        """
        if self.can_undo():
            self._now_index -= 1
            return self._ur_stack[self._now_index].clone()
        return None

    def redo(self) -> Optional[CloneableFile]:
        """
        (en) Returns the next pushed data.
        (ja) １つ後にpushされたデータを返します。
        """
        if self.can_redo():
            self._now_index += 1
            return self._ur_stack[self._now_index].clone()
        return None

    def now(self) -> Optional[CloneableFile]:
        """
        (en) Returns the now data.
        (ja) 現在のデータを返します。
        """
        if 0 <= self._now_index < len(self._ur_stack):
            return self._ur_stack[self._now_index].clone()
        return None

    def now_index(self) -> int:
        """
        (en) Returns the current index into the stack.
        (ja) 現在スタックのどこを参照しているのかという、インデックスを返します。
        """
        return self._now_index

    def get_stack(self) -> List[CloneableFile]:
        """
        (en) Returns a reference to the stack maintained by this class.
        (ja) このクラスで保持しているスタックの参照を返します。
        """
        return self._ur_stack
